"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

__NAME__ = "kimi"
__DESC__ = "metatutu library"
__VERSION__ = "0.0.5"
__LICENSE__= "MIT License"
__AUTHOR__ = "Wooloo Studio"
__AUTHOR_EMAIL__ = "max.wu@wooloostudio.com"
__URL__ = "https://github.com/metatutu/metatutu"
__PROJECT_URLS__ = {
	"Documentation": "https://github.com/metatutu/metatutu",
	"Source Code": "https://github.com/metatutu/metatutu",
	"Bug Tracker": "https://github.com/metatutu/metatutu/issues",
	"Forum": "https://github.com/metatutu/metatutu/issues",
	"Donate": "https://github.com/metatutu/metatutu"
}
__BUILD__ = 78